/*******************************************************************************
 * Copyright (C) 2012 Florian Feldbauer <florian@ep1.ruhr-uni-bochum.de>
 *                    - Ruhr-Universitaet Bochum, Lehrstuhl fuer Experimentalphysik I
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Brief: Description of temperature/resistance curves for different sensors
 *        connected to the THMP
 *
 * version 2.0.0; Dec. 13, 2013
 ******************************************************************************/

/**
 * @file thmp.c
 * @author F.Feldbauer
 * @date 13 Dec 2013
 * @brief Subroutine implementation for calibration of THMP
 */

/*_____ I N C L U D E S ______________________________________________________*/

/* ANSI C/C++ includes */
#include <gsl/gsl_poly.h>
#include <math.h>
#include <stdio.h>

/* EPICS includes */
#include <dbDefs.h>
#include <epicsExport.h>
#include <registryFunction.h>
#include <subRecord.h>

/*_____ D E F I N I T I O N S ________________________________________________*/

/* define NotANumber symbol if not existing */
#ifndef NAN
#  define NAN sqrt(-1)
#endif

/*_____ G L O B A L S ________________________________________________________*/
int thmpDebug = 0;

/*_____ L O C A L S __________________________________________________________*/
static const double t0 = 273.15;

/*_____ F U N C T I O N S ____________________________________________________*/

/**-----------------------------------------------------------------------------
 * @brief   Calibrate RTD sensors
 *
 * Calibration of RTD sensors using DIN norm like funciton
 * x(T) = a + bx + cx^2 + dx^3 where x is the conversion result of the ADC.
 * To determine the temperature the roots of this functions are calculated.
 * If more than one root is found, a linear approach is used to find best root.
 * The conversion result is stored in field E of the record.
 * Calibration parameters are stored in the fields A,B,C,D of the record.
 *
 * @param   [in]  prec   Address of the subroutine record calling this function
 *
 * @return  In case of error return -1, otherwise return 2 (no conversion)
 *----------------------------------------------------------------------------*/
static long thmpSolveCubic( subRecord *prec ) {

  if ( 0 == prec->d ) {
    fprintf ( stderr, "ERROR in thmpSolveCubic: Invalid value for field D in record %s (%p)\n",
              prec->name, (void*) prec );
    prec->val = NAN;
    return -1;
  }

  double a      = prec->c / prec->d;
  double b      = prec->b / prec->d;
  double c      = (prec->a - prec->e) / prec->d;
  double mean   = prec->f * ( prec->e - prec->a );
  double temp   = 100000.;
  double buffer = -t0;
  double x[3]   = { 0., 0., 0. };
  double dx[3]  = { 0., 0., 0. };
  int i         = 0;

  int numRoots  = gsl_poly_solve_cubic ( a, b, c, &x[0], &x[1], &x[2] );

  if ( thmpDebug ) {
    printf ( "Record %s called THMPsolveCubic (%p)\n", prec->name, (void*) prec );
    printf ( "  a = %f\n  b = %f\n  c = %f\n  mean = %f\n", a, b, c, mean );
    printf ( "  Found %d real root(s):\n", numRoots );
    for ( i = 0; i < numRoots; i++ )
      printf ( "    x%d = %f\n", i, x[i] );
  }

  if ( 1 == numRoots ) {

    /* Only one solution found */
    prec->val = x[0];

  } else {

    /* Found more than one solution: Determine best root */
    for ( i = 0; i < numRoots; i++ ) {
      dx[i] = fabs( mean - x[i] );
      if ( temp > dx[i] ) {
        temp = dx[i];
        buffer = x[i];
      }
    }
    prec->val = buffer;

  }

  return 0;
}

epicsExportAddress( int, thmpDebug );
epicsRegisterFunction( thmpSolveCubic );
